class LoggingConfig:
    """
    options:
        show_source: bool[true]
            add source info (filepath and line number) prefix to log messages.
            example:
                lk.log('hello world')
                # enabled : './main.py:10  >>  hello world'
                # disabled: 'hello world'
        show_varnames: bool[false]
            show both var names and values. (magic reflection)
            example:
                a, b = 1, 2
                lk.log(a, b, a + b)
                # enabled : 'main.py:11  >>  a = 1; b = 2; a + b = 3'
                # disabled: 'main.py:11  >>  1, 2, 3'
        show_external_lib: bool[true]
            if `param source` came from an external library, whether to print.
            for example, if a third-party library 'xxx' also used `lk.log`,
            its source path (relative to current working dir) may be very long,
            if you don't want to see any prints except your own project, you'd
            set this to False.

        # the following options are only available if `show_external_lib` is
        # true.
        path_format_for_external_lib: literal
            literal:
                'pretty_relpath': default
                    trunscate the source path of external lib to be shorter.
                    example:
                        before:
                            '../../../../site-packages/lk_logger/sourcemap.py'
                            # there may be a lot of '../'.
                        after:
                            '[lk_logger]/sourcemap.py'
                'relpath':
                    a relative path to current working dir. (<- `os.getcwd()`)
                    note there may be a lot of '../../../...' if external lib
                    is far beyond the current working dir.
                'lib_name_only':
                    show only the library name (surrounded by brackets).
                    example: '[lk_logger]'
            ps: if you don't want to show anything, you should turn to set
            `show_external_lib` to False.
    """
    show_source = True
    show_varnames = False
    show_external_lib = True
    path_format_for_external_lib = 'pretty_relpath'
    separator = ';   '  # suggests: ';   ' | ';\t' | '    ' | ...
    
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(LoggingConfig, k):
                setattr(self, k, v)
    
    def update(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
    
    def reset(self):
        self.show_source = True
        self.show_varnames = False
        self.show_external_lib = True
        self.path_format_for_external_lib = 'pretty_relpath'



from rich.console import Console as BaseConsole

__all__ = ['con_print', 'con_error']


class Console(BaseConsole):
    
    def __init__(self):
        super().__init__(soft_wrap=True)
        if self._color_system is None:
            self._color_system = 'standard'
        # TODO (width):
        #   if width longer than default, use single line style; otherwise
        #   split sourcemap and message into different lines.
        pass


console = Console()
con_print = console.print
con_error = console.print_exception



"""
index based text scanner.
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from enum import auto
import typing as t


class PairType:
    @dataclass
    class Plain:
        text: str
        symbol: tuple = ()
        children: list | tuple = ()
    
    @dataclass
    class Pair:
        text: str
        symbol: tuple[str, str]
        children: list | tuple


# noinspection PyArgumentList
class PairRule(Enum):
    ESCAPE = auto()  # e.g. '\\'
    COMMENT = auto()  # e.g. '#'
    PAIR_AB = auto()  # e.g. '()', '[]', '{}'
    PAIR_AA = auto()  # e.g. '""', "''", '``'
    BLOCK_AA = auto()  # e.g. '```...```'
    BLOCK_AB = auto()  # e.g. '/* ... */'


class T:
    _Start = str
    _End = str
    _Rule = PairRule
    PairRule = t.Iterable[tuple[tuple[_Start, _End], _Rule]]
    
    _Types = PairType
    PairType = t.Union[_Types.Plain, _Types.Pair]


BUILTIN_PAIR_RULES: T.PairRule = (
    (('\\', ''), PairRule.ESCAPE),
    # (('#', ''), PairRule.COMMENT),
    # (('//', ''), PairRule.COMMENT),
    (('(', ')'), PairRule.PAIR_AB),
    (('[', ']'), PairRule.PAIR_AB),
    (('{', '}'), PairRule.PAIR_AB),
    (('`', '`'), PairRule.BLOCK_AA),
    (('"', '"'), PairRule.PAIR_AA),
    (("'", "'"), PairRule.PAIR_AA),
    (('/*', '*/'), PairRule.BLOCK_AB),
    (('"""', '"""'), PairRule.BLOCK_AA),
    (("'''", "'''"), PairRule.BLOCK_AA),
)


class Scanner:
    
    def __init__(self, pair_rules: T.PairRule = BUILTIN_PAIR_RULES):
        from collections import defaultdict
        self._rules = {x[0][0]: x[1] for x in pair_rules}
        #   dict[str start, rule]
        self._pairs = dict(x[0] for x in pair_rules)
        #   dict[str start, str end]
        self._pairs_r = {v: k for k, v in self._pairs.items()}
        assert len(self._pairs) == len(self._pairs_r), (
            'text scanner does not support ambiguous pair rules, for example: '
            '`[` matches `]` while also `{` matches `]`!'
        )
        self._pair_initials = defaultdict(lambda : {
            'possible_symbols': [],
            'possible_rules': [],
        })
        #   {str initial_character: [rule, ...], ...}
        self._pair_starts = defaultdict(list)  # delete?
        self._pair_ends = defaultdict(list)  # delete?
        
        for (start, end), rule in pair_rules:
            self._pair_initials[start[0]]['possible_symbols'].append(start)
            self._pair_initials[start[0]]['possible_rules'].append(rule)
            for i in range(len(start)):
                self._pair_starts[start[:i + 1]].append(rule)
            for i in range(len(end)):
                self._pair_ends[end[:i + 1]].append(rule)
        ''' e.g.
            self._pair_starts = {
                '"': [PairRule.PAIR_AA, PairRule.BLOCK_AA],
                '""': [PairRule.BLOCK_AA],
                '"""': [PairRule.BLOCK_AA],
                ...
            }
        '''
    
    def scan(self, text: str):
        consumed = ''
        ctx = Context()
        goto_colx: int = 0
        pairs_scope = {}  # dict[str start, tuple[int colx_start, int colx_end]]
        wait_list = {
            'stack': [],  # list[str expect_end]
            #   the latest is in the first place. (we will use `insert` to add
            #   element into it.)
            'related': {},  # dict[str expect_end, int colx_of_start]
        }
        
        for rowx, line in enumerate(text.splitlines()):
            ctx.feed_line(line)
            
            for colx, char in enumerate(line):
                if colx < goto_colx:
                    continue
                if ctx.stage == ContextStage.PLAIN:
                    if x := self._pair_initials.get(char):
                        final_symbol = ''
                        for symbol in sorted(x['possible_symbols']):
                            if len(symbol) == 1:
                                final_symbol = symbol
                            else:
                                if line[colx : colx + len(symbol)] == symbol:
                                    final_symbol = symbol
                        if final_symbol:
                            rule = self._rules[final_symbol]
                            if rule == PairRule.ESCAPE:
                                goto_colx = colx + 1
                                continue
                            elif rule == PairRule.COMMENT:
                                goto_colx = len(line)
                                continue
                            elif rule == PairRule.PAIR_AB:
                                pass
                            
                            
                            ctx.stage = ContextStage.PAIR_START_DONE
                            symbol_end = self._pairs[final_symbol]
                            wait_list['stack'].insert(0, symbol_end)
                            wait_list['related'][symbol_end] = colx
                            goto_colx = colx + len(final_symbol)
                            continue
                        else:
                            consumed += char
                            continue
                    else:
                        consumed += char
                        continue
                elif ctx.stage == ContextStage.PAIR_START_DONE:
                    assert wait_list['stack']
                    for expect_end in wait_list['stack']:
                        pass
                    

# noinspection PyArgumentList
class ContextStage(Enum):
    PLAIN = auto()
    PAIRING_START = auto()
    PAIR_START_DONE = auto()
    PAIRING_END = auto()


class Context:
    stage: ContextStage = ContextStage.PLAIN
    start: str
    end: str
    _line: str
    
    def feed_line(self, line: str):
        self._line = line



""" WIP
a general purpose text scanner for extracting paired snippets from text.

example:
    input:
        aaa (bbb) "ccc 'ddd' (eee)" [fff [ggg]
    output (this is a conception, not the real result):
        [
            <Plain text='aaa ', symbol='', children=[]>,
            <SinglePair text='bbb', symbol='()', children=[]>,
            <MultiplePairs text='ccc \'ddd\' (eee)', symbol='""', children=[
                <Plain text='ccc ', symbol='', children=[]>,
                <SinglePair text='ddd', symbol='\'\'', children=[]>,
                <Plain text=' ', symbol='', children=[]>,
                <SinglePair text='eee', symbol='()', children=[]>,
            ]>,
            <Plain text=' [fff', symbol='', children=[]>,
            <MultiplePairs text='ggg', symbol='[]', children=[]>,
        ]
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from enum import auto
import typing as t


class PairType:
    @dataclass
    class Plain:
        text: str
        symbol: tuple = ()
        children: list | tuple = ()
    
    @dataclass
    class Pair:
        text: str
        symbol: tuple[str, str]
        children: list | tuple
    
    @dataclass
    class SinglePair:  # delete?
        text: str
        symbol: tuple[str, str]
        children: list | tuple = ()

    @dataclass
    class MultiplePairs:  # delete?
        text: str
        symbol: tuple[str, str]
        children: list | tuple


# noinspection PyArgumentList
class PairRule(Enum):
    ESCAPE = auto()  # e.g. '\\'
    COMMENT = auto()  # e.g. '#'
    PAIR_AB = auto()  # e.g. '()', '[]', '{}'
    PAIR_AA = auto()  # e.g. '""', "''", '``'
    BLOCK_AA = auto()  # e.g. '```...```'
    BLOCK_AB = auto()  # e.g. '/* ... */'


class T:
    _Start = str
    _End = str
    _Rule = PairRule
    PairRule = t.Iterable[tuple[tuple[_Start, _End], _Rule]]
    
    _Types = PairType
    PairType = t.Union[_Types.Plain, _Types.SinglePair, _Types.MultiplePairs]


BUILTIN_PAIR_RULES: T.PairRule = (
    (('\\', ''), PairRule.ESCAPE),
    # (('#', ''), PairRule.COMMENT),
    # (('//', ''), PairRule.COMMENT),
    (('(', ')'), PairRule.PAIR_AB),
    (('[', ']'), PairRule.PAIR_AB),
    (('{', '}'), PairRule.PAIR_AB),
    (('`', '`'), PairRule.BLOCK_AA),
    (('"', '"'), PairRule.PAIR_AA),
    (("'", "'"), PairRule.PAIR_AA),
    (('/*', '*/'), PairRule.BLOCK_AB),
    (('"""', '"""'), PairRule.BLOCK_AA),
    (("'''", "'''"), PairRule.BLOCK_AA),
)


class Scanner:
    
    def __init__(self, pair_rules: T.PairRule = BUILTIN_PAIR_RULES):
        from collections import defaultdict
        self._rules = pair_rules
        self._pairs = dict(x[0] for x in pair_rules)
        self._pairs_r = {v: k for k, v in self._pairs.items()}
        assert len(self._pairs) == len(self._pairs_r), (
            'text scanner does not support ambiguous pair rules, for example: '
            '`[` matches `]` while also `{` matches `]`!'
        )
        self._pair_starts = defaultdict(list)
        self._pair_ends = defaultdict(list)
        
        for (start, end), rule in pair_rules:
            for i in range(len(start)):
                self._pair_starts[start[:i + 1]].append(rule)
            for i in range(len(end)):
                self._pair_ends[end[:i + 1]].append(rule)
        ''' e.g.
            self._pair_starts = {
                '"': [PairRule.PAIR_AA, PairRule.BLOCK_AA],
                '""': [PairRule.BLOCK_AA],
                '"""': [PairRule.BLOCK_AA],
                ...
            }
        '''
        
    def scan(self, text: str):
        for line in text.splitlines():
            yield from self.scan_snippet(line)

    def scan_snippet(self, snip: str) -> t.Iterator[T.PairType]:
        """
        a snippet is a piece of text, which is a single line.
        this method can be recursively called.
        """
        consumed = ''
        context = Context()
        
        possibles: list[T.PairRule]
        determined: T.PairRule
        
        pairing_pointer = 0  # del
        pairing_initial = ''  # del
        pairing_symbol = ''  # del
        pair_start_store = ''
        pair_end_store = ''
        pair_end_expect = ''
        
        def is_complete(symbol: str, is_start: bool = True) -> bool:
            if is_start:
                return symbol in self._pairs
            else:
                return symbol in self._pairs_r
        
        for char in snip:
            if context.stage == ContextStage.PLAIN:
                if char not in self._pair_starts:
                    consumed += char
                    continue
                else:
                    yield PairType.Plain(consumed)
                    consumed = ''
                    
                    context.stage = ContextStage.PAIRING_START
                    pair_start_store = ''  # reset
            
            if context.stage == ContextStage.PAIRING_START:
                pair_start_store += char
                possibles = self._pair_starts[pair_start_store]
                if len(possibles) == 1:
                    if is_complete(pair_start_store):
                        # the pair is complete.
                        context.rule = possibles[0]
                        context.stage = ContextStage.PAIR_START_DONE
                        pair_end_expect = self._pairs[pair_start_store]
                        pair_end_store = ''  # reset
                        continue
                    else:
                        continue
                pass
            
            if context.stage == ContextStage.PAIR_START_DONE:
                pair_end_store += char
                if pair_end_store in self._pairs_r:
                    # the pair match is complete.
                    yield PairType.SinglePair()
                if context.rule == PairRule.PAIR_AB:
                    end = self._pairs[pair_start_store]
                    
                if char in self._pair_starts:
                    pass
                
                if char not in self._pair_starts:
                    consumed += char
                    continue


# noinspection PyArgumentList
class ContextStage(Enum):
    PLAIN = auto()
    PAIRING_START = auto()
    PAIR_START_DONE = auto()
    PAIRING_END = auto()
    

class Context:
    stage: ContextStage = ContextStage.PLAIN
    rule: t.Optional[T.PairRule] = None
    # pair_store: str = ''
    start: str = ''
    end: str = ''
